"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Groove = exports.CacheModeEnum = void 0;
const util_1 = require("util");
const child_process_1 = require("child_process");
const promises_1 = require("fs/promises");
const fs_1 = require("fs");
const path_1 = require("path");
const utilities_1 = require("./utilities");
const tape_1 = require("./tape");
const os_1 = require("os");
const form_data_1 = __importDefault(require("form-data"));
exports.CacheModeEnum = {
    OFF: 0,
    STANDARD: 1,
    AGGRESSIVE_GET: 2,
    AGGRESSIVE: 3
};
const checkStatus = async (response, echoError) => {
    if (response.status > 300 || response.status < 200) {
        console.log(`Error: ${response.status}: ${await response.text()}`);
        throw Error(echoError);
    }
    const contents = await response.json();
    if (contents["success"] != true) {
        throw Error(echoError);
    }
};
class Groove {
    constructor(config) {
        this.process = null;
        this.executablePath = null;
        this.commandTimeout = config.commandTimeout || 5000;
        this.port = config.port || 6010;
        this.controlPort = config.controlPort || 6011;
        this.authUsername = config.authUsername || null;
        this.authPassword = config.authPassword || null;
        this.baseUrlProxy = `http://localhost:${this.port}`;
        this.baseUrlControl = `http://localhost:${this.controlPort}`;
        this.certificate = (0, fs_1.readFileSync)((0, path_1.join)((0, os_1.homedir)(), '.grooveproxy/ca.crt'));
        this.silenceLogs = false;
        this.launched = false;
    }
    async launch() {
        if (this.launched) {
            throw Error("Already spawned proxy.");
        }
        const parameters = {
            "--port": this.port ? this.port.toString() : null,
            "--control-port": this.controlPort ? this.controlPort.toString() : null,
            "--auth-username": this.authUsername,
            "--auth-password": this.authPassword,
        };
        const exc = await this.getExecutablePath();
        this.process = (0, child_process_1.spawn)(exc, Object.entries(parameters).reduce((previous, [key, value]) => {
            if (value != null) {
                previous = [...previous, key, value];
            }
            return previous;
        }, []));
        this.launched = true;
        this.process.stdout.setEncoding('utf8');
        this.process.stdout.on('data', (data) => {
            if (this.silenceLogs)
                return;
            console.log(`proxy-stdout: ${data}`);
        });
        this.process.stderr.setEncoding('utf8');
        this.process.stderr.on('data', (data) => {
            if (this.silenceLogs)
                return;
            console.log(`proxy-stderr: ${data}`);
        });
        this.process.on('close', (code) => {
            if (this.silenceLogs)
                return;
            console.log(`child process exited with code ${code}`);
        });
        // Ensure we cleanup after ourselves
        process.on('exit', () => {
            console.log("Will quit");
            this.stop();
        });
        await (0, utilities_1.sleep)(1000);
    }
    async stop() {
        this.silenceLogs = true;
        if (this.process)
            this.process.kill();
        this.launched = false;
        this.process = null;
    }
    async tapeStart() {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/record`, {
            method: "POST",
            timeout: this.commandTimeout,
        });
        await checkStatus(response, "Failed to start recording.");
    }
    async tapeGet(tapeID) {
        tapeID = tapeID || "";
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/retrieve`, {
            method: "POST",
            timeout: this.commandTimeout,
            body: JSON.stringify({
                tapeID,
            }),
        });
        const contents = response.body;
        const session = new tape_1.TapeSession();
        session.readFromServer(await (0, utilities_1.streamToBuffer)(contents));
        return session;
    }
    async tapeLoad(session) {
        const formData = new form_data_1.default();
        // Filename is irrelevant but required to upload payload buffer as file
        formData.append("file", session.toServer(), { filename: 'tape.json.gzip' });
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/load`, {
            method: "POST",
            timeout: this.commandTimeout,
            body: formData,
        });
        await checkStatus(response, "Failed to load tape.");
    }
    async tapeStop() {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/stop`, {
            method: "POST",
            timeout: this.commandTimeout,
        });
        await checkStatus(response, "Failed to stop recording.");
    }
    async tapeClear(tapeID) {
        tapeID = tapeID || "";
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/clear`, {
            method: "POST",
            timeout: this.commandTimeout,
            body: JSON.stringify({
                tapeID,
            }),
        });
        await checkStatus(response, "Failed to clear tape.");
    }
    async setCacheMode(mode) {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/cache/mode`, {
            method: "POST",
            timeout: this.commandTimeout,
            body: JSON.stringify({ mode }),
        });
        await checkStatus(response, "Failed to set cache mode.");
    }
    async cacheClear() {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/cache/clear`, {
            method: "POST",
            timeout: this.commandTimeout,
        });
        await checkStatus(response, "Failed to clear cache.");
    }
    async dialerLoad(dialers) {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/dialer/load`, {
            method: "POST",
            timeout: this.commandTimeout,
            body: JSON.stringify({
                definitions: dialers.map((dialer) => ({
                    priority: dialer.priority,
                    proxyServer: dialer.proxy ? dialer.proxy.url : null,
                    proxyUsername: dialer.proxy ? dialer.proxy.username : null,
                    proxyPassword: dialer.proxy ? dialer.proxy.password : null,
                    requiresUrlRegex: dialer.requestRequires ? dialer.requestRequires.urlRegex : null,
                    requiresResourceTypes: dialer.requestRequires ? dialer.requestRequires.resourceTypes : null,
                }))
            }),
        });
        await checkStatus(response, "Failed to start end-proxy.");
    }
    async endProxyStop() {
        const response = await (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/proxy/stop`, {
            method: "POST",
            timeout: this.commandTimeout,
        });
        await checkStatus(response, "Failed to stop end-proxy.");
    }
    async getExecutablePath() {
        if (this.executablePath)
            return this.executablePath;
        const npmBin = await (0, util_1.promisify)(child_process_1.exec)("npm bin");
        if (npmBin.stderr || !npmBin.stdout) {
            throw Error("Unknown grooveproxy executable location");
        }
        const binDirectory = npmBin.stdout.trim();
        this.executablePath = (0, path_1.join)(binDirectory, "grooveproxy");
        // Determine if the path exists, will raise an error if not
        await (0, promises_1.stat)(this.executablePath);
        return this.executablePath;
    }
}
exports.Groove = Groove;
//# sourceMappingURL=proxy.js.map